import logging
from abc import abstractmethod
from copy import deepcopy
from typing import Dict, Type, Callable, Union, Iterable
from wcmatch import glob

import numpy as np
import torch as t
from torch import nn

from pandas_ml_common.utils.logging_utils import LogOnce
from pandas_ml_utils import AutoEncoderModel, call_callable_dynamic_args
from pandas_ml_utils_torch.utils import to_device

_log = logging.getLogger(__name__)


class PytorchNN(nn.Module):

    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, *input, state=None, **kwargs):
        if self.training:
            return self.forward_training(*input)
        else:
            if state == AutoEncoderModel.ENCODE:
                return self.encode(*input)
            elif state == AutoEncoderModel.DECODE:
                return self.decode(*input)
            else:
                return self.forward_predict(*input)

    @abstractmethod
    def forward_training(self, *input) -> t.Tensor:
        pass

    def forward_predict(self, *input) -> t.Tensor:
        return self.forward_training(*input)

    def encode(self, *input) -> t.Tensor:
        raise NotImplementedError("For autoencoders the methods `encode` and `decode` need to be implemented!")

    def decode(self, *input) -> t.Tensor:
        raise NotImplementedError("For autoencoders the methods `encode` and `decode` need to be implemented!")

    def L1(self) -> Dict[str, float]:
        return {}

    def L2(self) -> Dict[str, float]:
        return {}


class PytochBaseModel(object):

    _MODEL_FIELD = "net"

    def __init__(self,
                 net_provider: Union[Type[PytorchNN], Callable[[], PytorchNN]],
                 criterion_provider: Union[Type[nn.modules.loss._Loss], Callable[[], nn.modules.loss._Loss]],
                 optimizer_provider: Union[Type[t.optim.Optimizer], Callable[[Iterable], t.optim.Optimizer]],
                 record_best_weights: bool = False,
                 cuda: bool = False
                 ):

        self.net_provider = net_provider
        self.criterion_provider = criterion_provider
        self.optimizer_provider = optimizer_provider

        self.net = to_device(net_provider(), cuda)
        self.criterion = to_device(call_callable_dynamic_args(criterion_provider, module=self.net, params=self.net.named_parameters()), cuda)
        self.optimizer = optimizer_provider(self.net.parameters())
        self.log_once = LogOnce().log
        self.best_weights = None
        self.best_loss = float('inf')
        self.last_loss = float('inf')
        self.record_best_weights = record_best_weights

        # initialization
        self.optimizer.zero_grad()
        if hasattr(self.net, "init_weights"):
            self.net.apply(self.net.init_weights)

        # penalty
        self.l1_penalty_tensors = {t.zeros(1): 1.}
        self.l2_penalty_tensors = {t.zeros(1): 1.}

        def param_dict(module):
            return {t[0].replace('.', '/'): t[1] for t in module.named_parameters()}

        if hasattr(self.net, "L1") and len(self.net.L1()) > 0:
            self.l1_penalty_tensors = \
                {tensor: penalty for param, tensor in param_dict(self.net).items()
                 for path, penalty in self.net.L1().items()
                 if glob.globmatch(param, path, flags=glob.GLOBSTAR)}

        if hasattr(self.net, "L2") and len(self.net.L2()) > 0:
            self.l2_penalty_tensors = \
                {tensor: penalty for param, tensor in param_dict(self.net).items()
                 for path, penalty in self.net.L2().items()
                 if glob.globmatch(param, path, flags=glob.GLOBSTAR)}

    def cpu(self):
        self.net = self.net.cpu()
        self.criterion = self.criterion.cpu()
        return self

    def cuda(self):
        self.net = self.net.cuda()
        self.criterion = self.criterion.cuda()
        return self

    def fit_epoch(self, x: t.Tensor, y: t.Tensor, sample_weight: t.Tensor = None):
        if not self.net.training:
            self.net.train()

        # ===================forward=====================
        output = self.net(x)
        loss = self._calc_weighted_loss(self.criterion, output, y, sample_weight)

        # ===================backward====================
        loss.backward()
        self.optimizer.step()
        self.optimizer.zero_grad()

        #
        loss_value = loss.cpu().item()
        self.last_loss = loss_value
        if loss_value < self.best_loss:
            self.best_loss = loss_value
            if self.record_best_weights:
                self.best_weights = deepcopy(self.net.state_dict())

        # print('lala', loss_value)
        return loss_value

    def calculate_loss(self, x: t.Tensor, y_true: t.Tensor, sample_weight: t.Tensor):
        with t.no_grad():
            y_pred = self.net(x)
            return self._calc_weighted_loss(self.criterion, y_pred, y_true, sample_weight).cpu().item()

    def _calc_weighted_loss(self, criterion: nn.modules.loss._Loss, y_hat: t.Tensor, y: t.Tensor, weights: t.Tensor):
        l1 = t.stack([penalty * tensor.norm(p=1) for tensor, penalty in self.l1_penalty_tensors.items()]).sum()
        l2 = t.stack([penalty * tensor.norm(p=2) ** 2 for tensor, penalty in self.l2_penalty_tensors.items()]).sum()
        loss = criterion(y_hat, y) + l1 + l2

        if loss.ndim > 0:
            if weights is None:
                weights = t.ones(y_hat.shape[0])

            if loss.ndim == weights.ndim:
                loss = (loss * weights).mean()
            elif weights.ndim > loss.ndim and weights.shape[-1] == 1:
                loss = (loss * weights.squeeze()).mean()
            else:
                self.log_once("loss.ndim!=weights.ndim", _log.warning,
                              f"sample weight has different dimensions {loss.shape}, {weights.shape}")

                loss = (loss * weights.repeat(1, *loss.shape[1:])).mean()

        return loss

    def restore_best_weights(self):
        if self.best_weights is not None:
            self.net.load_state_dict(self.best_weights)
        else:
            _log.warning("No best weights found!! Keep existing weights.")

    def predict(self, x: t.Tensor, samples=1, numpy=True, force_mode: bool = False) -> Union[np.ndarray, t.Tensor]:
        if force_mode and self.net.training:
            self.net.eval()

        with t.no_grad():
                y_hat = t.stack([self.net(x) for _ in range(samples)], dim=1) if samples > 1 else self.net(x)

        return y_hat.cpu().numpy() if numpy else y_hat

    def train(self):
        self.net.train()
        return self

    def eval(self):
        self.net.eval()
        return self

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains all our instance attributes.
        # Always use the dict.copy() method to avoid modifying the original state.
        state = self.__dict__.copy()

        # remove un-pickleable fields
        del state[PytochBaseModel._MODEL_FIELD]

        # add torch serialisation
        state[f'{PytochBaseModel._MODEL_FIELD}_state_dict'] = self.net.state_dict() if self.net is not None else None

        # return altered state
        return state

    def __setstate__(self, state):
        # use torch.save(model.state_dict(), './sim_autoencoder.pth')
        # first remove the special state
        module_state_dict = state[f'{PytochBaseModel._MODEL_FIELD}_state_dict']
        del state[f'{PytochBaseModel._MODEL_FIELD}_state_dict']

        # Restore instance attributes
        self.__dict__.update(state)
        self.net = self.net_provider()

        # restore special state dict
        if module_state_dict is not None:
            self.net.load_state_dict(module_state_dict)

        # restore optimizer and loss, this allows us to continue training
        self.optimizer = self.optimizer_provider(self.net.parameters())
        self.criterion = to_device(
            call_callable_dynamic_args(self.criterion_provider, module=self.net, params=self.net.named_parameters()),
            False
        )



import logging
import sys
from typing import Dict

from pandas_ml_utils_torch.pytorch_base import PytochBaseModel

_log = logging.getLogger(__name__)


def take_the_best(folds: Dict[int, PytochBaseModel]):
    best_loss = sys.float_info.max
    best_module = None
    best_fold = -1

    for fold, model in folds.items():
        _log.info(f'fold: {fold}, loss: {model.last_loss}, best loss: {model.best_loss}')
        if model.last_loss < best_loss:
            best_module = model
            best_fold = fold

    _log.info(f'best fold: {best_fold}!')
    return best_module






from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from functools import wraps
from typing import List, Callable, Type, Dict

import pandas as pd
import torch as t
import torch.nn as nn

from pandas_ml_common import Typing
from pandas_ml_common.utils import to_pandas
from pandas_ml_common.utils.logging_utils import LogOnce
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.model.base_model import Model, AutoEncoderModel
from pandas_ml_utils.ml.summary import Summary
from pandas_ml_utils_torch.pytorch_base import PytorchNN, PytochBaseModel
from pandas_ml_utils_torch.utils import from_pandas

_log = logging.getLogger(__name__)


class _AbstractPytorchModel(Model):

    def __init__(self,
                 module_provider: Type[PytorchNN],
                 features_and_labels: FeaturesAndLabels,
                 criterion_provider: Type[nn.modules.loss._Loss],
                 optimizer_provider: Type[t.optim.Optimizer],
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 restore_best_weights: bool = False,
                 merge_cross_folds: Callable[[Dict[int, PytochBaseModel]], PytochBaseModel] = None,
                 callbacks: Dict[str, List[Callable]] = {},
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.module_provider = module_provider
        self.merge_cross_folds = merge_cross_folds
        self.callbacks = callbacks

        self.log_once = LogOnce().log
        self._current_model: PytochBaseModel = module_provider()
        self._cuda = False

        self.models: Dict[int, PytochBaseModel] = defaultdict(
            lambda: PytochBaseModel(lambda: deepcopy(self._current_model), criterion_provider, optimizer_provider, restore_best_weights, self._cuda))

    def init_fit(self, **kwargs):
        self._cuda = kwargs.get("cuda", False)

    def init_fold(self, epoch: int, fold: int):
        self._current_model = self.models[fold]

    def fit_batch(self, x: pd.DataFrame, y: pd.DataFrame, weight: pd.DataFrame, fold: int, **kwargs):
        self._current_model.fit_epoch(from_pandas(x, self._cuda), from_pandas(y, self._cuda), from_pandas(weight, self._cuda))

    def calculate_loss(self, fold, x, y_true, weight):
        # return self._current_model.calculate_loss(
        return self.models[fold].calculate_loss(
            from_pandas(x, self._cuda),
            from_pandas(y_true, self._cuda),
            from_pandas(weight, self._cuda),
        )

    def merge_folds(self, epoch: int):
        if len(self.models) > 1:
            if self.merge_cross_folds is not None:
                self._current_model = self.merge_cross_folds(self.models)
                self.models.clear()
            else:
                self.log_once("merge_folds", _log.warning,
                              f"no merge folds function suplied, keep training {len(self.models)} independent")

    def finish_learning(self):
        self.models.clear()

    def _predict(self, features: pd.DataFrame, col_names, samples=1, cuda=False, **kwargs) -> Typing.PatchedDataFrame:
        if cuda:
            self._current_model = self._current_model.cuda()
        else:
            self._current_model = self._current_model.cpu()

        return to_pandas(self._current_model.predict(from_pandas(features, cuda), samples), features.index, col_names)


class PytorchModel(_AbstractPytorchModel):

    @wraps(_AbstractPytorchModel.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def predict(self, features: pd.DataFrame, targets: pd.DataFrame = None, latent: pd.DataFrame = None, samples=1, cuda=False, **kwargs) -> Typing.PatchedDataFrame:
        self._current_model.eval()
        return self._predict(features, self._labels_columns, samples, cuda, **kwargs)


class PytorchAutoEncoderModel(_AbstractPytorchModel, AutoEncoderModel):

    @wraps(_AbstractPytorchModel.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _auto_encode(self, features: pd.DataFrame, samples, cuda=False, **kwargs) -> Typing.PatchedDataFrame:
        self._current_model.eval()
        return self._predict(features, self._labels_columns, samples, cuda, **kwargs)

    def _encode(self, features: pd.DataFrame, samples, cuda=False, **kwargs) -> Typing.PatchedDataFrame:
        self._current_model.eval()
        return self._predict(features, self._features_and_labels.latent_names, samples, cuda, **kwargs)

    def _decode(self, latent_features: pd.DataFrame, samples, cuda=False, **kwargs) -> Typing.PatchedDataFrame:
        self._current_model.eval()
        return self._predict(latent_features,  self._labels_columns, samples, cuda, **kwargs)




from .pytorch_model import PytorchModel, PytorchAutoEncoderModel
from .pytorch_base import PytorchNN
from .layers import Reshape, KerasLikeLSTM, ResidualLayer, Flatten, Squeeze
from .loss import MultiObjectiveLoss, CrossEntropyLoss, HeteroscedasticityLoss



from typing import Tuple, Union

from pandas_ml_common import pd
import torch as t

from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_common.utils import unpack_nested_arrays


def from_pandas(df: pd.DataFrame, cuda: bool = False, default: t.Tensor = None) -> Union[t.Tensor, Tuple[t.Tensor, ...]]:
    if df is None and default is None:
        return None

    if isinstance(df, MultiFrameDecorator):
        return tuple([from_pandas(f, cuda, default) for f in df.frames()])

    val = (t.from_numpy(df._.get_values(split_multi_index_rows=False, squeeze=True)) if df is not None else default).float()
    return val.cuda() if cuda else val.cpu()


def to_device(var, cuda):
    return var.cuda() if cuda else var.cpu()





import numpy as np
import pandas as pd

from pandas_ml_common import dummy_splitter
from pandas_ml_utils import Model, FittingParameter
from pandas_ml_utils.constants import *

def create_line_data(n=300, slope=1):
    np.random.seed(32)
    x = np.linspace(0, 1, n)
    y = slope * x + np.random.normal(0, 0.05, n)
    return x, y


def create_sine_data(n=300):
    np.random.seed(32)
    x = np.linspace(0, 1 * 2 * np.pi, n)
    y1 = 3 * np.sin(x)
    y1 = np.concatenate((np.zeros(60), y1 + np.random.normal(0, 0.15 * np.abs(y1), n), np.zeros(60)))
    x = np.concatenate((np.linspace(-3, 0, 60), np.linspace(0, 3 * 2 * np.pi, n),
                        np.linspace(3 * 2 * np.pi, 3 * 2 * np.pi + 3, 60)))
    y2 = 0.1 * x + 1
    y = (y1 + y2) + 2
    return x, y


if __name__ == '__main__':

    df = pd.DataFrame(np.array(create_line_data(300)).T, columns=["x", "y"])

    with df.model() as m:
        from pandas_ml_utils.ml.callback.live_loss_plot import NbLiveLossPlot
        from pandas_ml_utils_torch import PytorchModel, PytorchNN
        from pandas_ml_utils import FeaturesAndLabels
        from torch.optim import Adam
        from torch import nn
        import torch as t

        class Net(PytorchNN):

            def __init__(self):
                super(Net, self).__init__()
                self.net = nn.Sequential(
                    nn.Linear(1, 200),
                    nn.ReLU(),
                    nn.Linear(200, 200),
                    nn.ReLU(),
                    nn.Linear(200, 1),
                    nn.ReLU()
                )

            def forward_training(self, *input) -> t.Tensor:
                return self.net(input[0])

        fit = m.fit(
            PytorchModel(Net, FeaturesAndLabels(["x"], ["y"]), nn.MSELoss, Adam),
            FittingParameter(splitter=dummy_splitter, epochs=100),
            callbacks=[NbLiveLossPlot(backend='pgf')]
        )







import torch as t

PI = t.acos(t.zeros(1)) * 2


class HeteroscedasticityLoss(t.nn.Module):

    def __init__(self, reduction=None):
        super().__init__()
        self.reduction = reduction

    def forward(self, y_pred, y_true):
        mu = t.narrow(y_pred, 1, 0, 1)
        sigma = t.exp(t.narrow(y_pred, 1, 1, 1))

        a = 1 / (t.sqrt(2 * PI) * sigma)
        b1 = t.square(mu - y_true)
        b2 = 2 * t.square(sigma)
        b = b1 / b2

        loss = (-t.log(a) + b).squeeze()
        if self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'mean':
            return loss.mean()
        else:
            return loss




import os
import tempfile
import uuid
from unittest import TestCase

import numpy as np
import torch as t
from torch import nn
from torch.optim import Adam

from pandas_ml_common import serialize, deserialize
from pandas_ml_utils_torch.pytorch_base import PytochBaseModel, PytorchNN


class ANet(PytorchNN):

    def __init__(self, something):
        super().__init__()
        self.net = nn.Linear(1, 1, False)
        print(something)

    def forward_training(self, *input) -> t.Tensor:
        return self.net(input[0])


class TestPytorchBaseModel(TestCase):

    def test_linreg_fit(self):
        x = t.from_numpy(np.linspace(0, 1, 300).reshape(-1, 1)).float()
        y = t.from_numpy((np.linspace(0, 1, 300) + np.random.normal(0, 0.05, 300)).reshape(-1, 1)).float()

        m = PytochBaseModel(
            lambda: ANet("ANet prints: haha"),
            nn.MSELoss,
            Adam
        )

        err = [m.fit_epoch(x, y, None) for _ in range(3000)]
        self.assertLess(err[-1], err[0])

        self.assertAlmostEqual(((m.predict(x, numpy=False) - y) ** 2).mean().sqrt().cpu().item(), 0.05, 1)
        # print(m.predict(x).numpy())
        # print(y.numpy())

        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            serialize(m, temp)
            self.assertLess(((m.predict(x, numpy=False) - deserialize(temp, PytochBaseModel).predict(x, numpy=False)) ** 2).mean().cpu().item(), 1e-5)
        finally:
            os.remove(temp)

    def test_non_linear_fit(self):
        def create_sine_data(n=300):
            np.random.seed(32)
            x = np.linspace(0, 1 * 2 * np.pi, n)
            y1 = 3 * np.sin(x)
            y1 = np.concatenate((np.zeros(60), y1 + np.random.normal(0, 0.15 * np.abs(y1), n), np.zeros(60)))
            x = np.concatenate((np.linspace(-3, 0, 60), np.linspace(0, 3 * 2 * np.pi, n),
                                np.linspace(3 * 2 * np.pi, 3 * 2 * np.pi + 3, 60)))
            y2 = 0.1 * x + 1
            y = (y1 + y2) + 2
            return t.from_numpy(x).reshape(-1, 1).float(), t.from_numpy(y).reshape(-1, 1).float()

        t.manual_seed(0)

        class Net(PytorchNN):

            def __init__(self):
                super().__init__()
                self.net = nn.Sequential(
                    nn.Linear(1, 200),
                    nn.ReLU(),
                    nn.Linear(200, 200),
                    nn.ReLU(),
                    nn.Linear(200, 1),
                    nn.ReLU(),
                )

            def forward_training(self, *input) -> t.Tensor:
                return self.net(input[0])

        m = PytochBaseModel(
            Net,
            nn.MSELoss,
            Adam
        )

        x, y = create_sine_data()
        err = [m.fit_epoch(x, y, None) for _ in range(3000)]
        self.assertLess(err[-1], err[0])

        dist = ((m.predict(x, numpy=False) - y) ** 2).mean().sqrt().cpu().item()
        self.assertLess(dist, 0.5)
        # print(dist)
        # print("x = np." + repr(y.numpy()).replace(', dtype=float32', ''))
        # print("y = np." + repr(m.predict(x).numpy()).replace(', dtype=float32', ''))



import torch as t
import torch.nn as nn

from pandas_ml_utils_torch.loss import CrossEntropyLoss


class DifferentiableArgmax(nn.Module):

    def __init__(self, nr_of_categories, beta=1e10):
        super().__init__()
        self.softmax = nn.Softmax(dim=-1)
        self.register_buffer('y_range', t.arange(0, nr_of_categories).float())
        self.register_buffer('beta', t.tensor(beta).float())

    def forward(self, y):
        return t.sum(self.softmax(y * self.beta) * self.y_range, dim=-1)


class ParabolicPenaltyLoss(nn.Module):

    def __init__(self, nr_of_categories, delta=1.0, beta=1e10):
        super().__init__()
        self.argmax = DifferentiableArgmax(nr_of_categories, beta)
        self.register_buffer('offset', t.tensor(delta / 2).float())
        self.register_buffer('f', t.tensor((nr_of_categories + delta) / nr_of_categories).float())

    def forward(self, y_pred, y_true):
        return ((self.argmax(y_pred) + self.offset) - (self.argmax(y_true)) * self.f) ** 2


class TailedCategoricalCrossentropyLoss(nn.Module):

    def __init__(self, nr_of_categories: int, alpha=0.1, beta=1e10, delta=1.0, reduction='none'):
        """
        assuming that we have discretized something like returns where we have less observations in the tails.
        If we want to train a neural net to place returns into the expected bucket we want to penalize if the
        prediction is too close to the mean. we rather want to be pessimistic and force the predictor to
        encounter the tails.

        :param nr_of_categories: number of categories aka length of the one hot encoded vectors
        :param alpha: describes the steepness of the parabola
        :param beta: used for the differentiable_argmax
        :param delta: is used to un-evenly skew the loss to the outer bounds. 0 now skew > bigger skew
        :return: returns a keras loss function
        """
        super().__init__()
        self.parabolic_penalty = ParabolicPenaltyLoss(nr_of_categories, delta, beta)
        self.categorical_crossentropy = CrossEntropyLoss()
        self.reduction = reduction
        self.alpha = alpha

    def forward(self, y_pred, y_true):
        y_true = y_true.view(y_pred.shape)
        penalty = self.alpha * self.parabolic_penalty(y_true, y_pred).squeeze()
        loss = self.categorical_crossentropy(y_pred, y_true)
        loss = penalty + loss

        if self.reduction == 'sum':
            return t.sum(loss)
        elif self.reduction == 'mean':
            return t.mean(loss)
        else:
            return loss.view(y_pred.shape[0])




from unittest import TestCase

import torch as t
import torch.nn as nn

from pandas_ml_common import np
from pandas_ml_common.utils.column_lagging_utils import lag_columns
from pandas_ml_utils_test.config import DF_TEST
from pandas_ml_utils_torch.layers import KerasLikeLSTM, Time2Vec


class TestPytorchLayer(TestCase):

    def test_lstm_layer(self):
        data = t.rand((1, 10, 2))
        lstm = KerasLikeLSTM((10, 2), 2, return_sequence=False)
        lstm2 = KerasLikeLSTM((10, 2), 2, return_sequence=True)

        self.assertEqual((1, 2), lstm(data).shape)
        self.assertEqual((1, 10, 2), lstm2(data).shape)

    def test_time2vec(self):
        df = DF_TEST.copy()

        x = lag_columns(df["Volume"].pct_change(), 5).dropna()._.values[:-2]
        y = df["Volume"].pct_change().shift(-1).dropna()[6:]

        class MyModule(nn.Module):

            def __init__(self, input, output):
                super().__init__()
                self.net = Time2Vec(input, output)

            def forward(self, x):
                return self.net(x)

        t2v = MyModule(5, 32)
        print(x.shape)
        res = t2v(t.from_numpy(x).float())
        print(res)
        print(res.shape)  # -> ?, 5, 33

    def _test_lala(self):
        class MyModule(nn.Module):

            def __init__(self, input, output):
                super().__init__()
                self.net = Time2Vec(input, output)

            def forward(self, x):
                return self.net(x)

        t2v = MyModule(3, 3)
        print(t2v(t.from_numpy(np.array([[[0.1], [0.2], [0.3]], [[0.1], [0.2], [0.3]]])).float()))